package org.apache.xindice.core;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: Database.java,v 1.17 2002/12/27 18:37:23 dviner Exp $
 */

import org.apache.xindice.core.query.QueryEngine;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.Named;
import org.apache.xindice.util.XindiceException;
import org.apache.xindice.core.meta.TimeRecord;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.w3c.dom.*;

import java.io.File;
import java.util.HashMap;
import java.util.Map;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

/**
 * Database is the primary container for the Xindice Database Engine.
 */

public final class Database extends Collection implements Named {
   public static final String DBROOT = "dbroot";
   private static final String QUERYENGINE = "queryengine";
   private static final String DATABASE = "database";
   private static final String COLKEY = "database.xml";
   private static final String METADATA = "use-metadata";
   public static final String NAME = "name";


   private static Log log = LogFactory.getLog("org.apache.xindice.core");

   public static final String PROP_XINDICE_HOME = "xindice.home";

   private static final Map databases = new HashMap(); // String to Database

   private DocumentCache docCache = new DocumentCache();

   private SystemCollection sysCol = null;

   private MetaSystemCollection metaCol = null;

   private boolean sysInit = false;
   private boolean metaInit = false;
   private boolean metaEnabled = false;
   // key = canonical name, value = timerecord
   private Map timestamps = new HashMap();

   private QueryEngine engine = new QueryEngine(this);

   public static Database getDatabase(String name) {
      return (Database)databases.get(name);
   }

   public static String[] listDatabases() {
      return (String[]) databases.keySet().toArray(new String[0]);
   }

   public void setConfig(Configuration config) {
      this.config = config;

      name = config.getAttribute(NAME);
      setCanonicalName('/' + getName());

      String dbroot = config.getAttribute(DBROOT);
      File dbrootDir = new File(dbroot);
      if (!dbrootDir.isAbsolute()) {
         dbrootDir = new File(System.getProperty( PROP_XINDICE_HOME ), dbroot);
      }
      setCollectionRoot(dbrootDir);
      if (log.isInfoEnabled()) {
         log.info("Database points to " + dbrootDir.getAbsolutePath());
      }

      try {
         Configuration queryCfg = config.getChild(QUERYENGINE);
         if ( queryCfg != null )
            engine.setConfig(queryCfg);
      }
      catch ( Exception e ) {
         if (log.isDebugEnabled()) {
            log.debug("No message", e);
         }
      }

      if ( !sysInit ) {
         sysCol = new SystemCollection(this);

         try {
            sysCol.init();
         }
         catch ( XindiceException e ) {
            if (log.isDebugEnabled()) {
               log.debug("No message", e);
            }
         }

         collections.put(sysCol.getName(), sysCol);
         sysInit = true;
      }

      try {
         // Bootstrap from the database itself...  This is accomplished
         // by intercepting the setConfig call and using a Configuration
         // retrieved from the database instead of the standard config
         Document colDoc = sysCol.getCollection(SystemCollection.CONFIGS).getDocument(COLKEY);
         if ( colDoc == null ) {
            DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            colDoc = db.newDocument();
            Element root = colDoc.createElement(DATABASE);
            root.setAttribute(NAME, name);
            colDoc.appendChild(root);
            sysCol.getCollection(SystemCollection.CONFIGS).setDocument(COLKEY, colDoc);
         }

         super.setConfig(new Configuration(colDoc.getDocumentElement(), false));
      }
      catch ( Exception e ) {
         if (log.isDebugEnabled()) {
            log.debug("No message", e);
         }
      }

      // Register the Database with the VM
      databases.put(getName(), this);

      // initialize the meta collection
      // but only if it's turned on in the config.
      String metaCfg = config.getAttribute(METADATA);
      if ( metaCfg.equalsIgnoreCase("on") )
      {
         metaEnabled = true;
         if (!metaInit)
         {
            metaCol = new MetaSystemCollection(this);
            try {
               metaCol.init();
            }
            catch ( XindiceException e ) {
               if (log.isDebugEnabled()) {
                  log.debug("Error initializing the meta collection", e);
               }
            }
            // should this attach the meta collection to the database?

            collections.put(metaCol.getName(), metaCol);
            metaInit = true;
            if (log.isInfoEnabled()) {
               log.info("Meta information initialized");
            }
         }
      }

      // observer
      DBObserver.getInstance().setDatabaseConfig(this, collections, config);
   }

   public SystemCollection getSystemCollection() {
      return sysCol;
   }

   /**
    * Return the MetaSystem collection for this database.
    * 
    * It will return null if metadata is not enabled on this database.
    * @return MetaSystemCollection
    */
   public MetaSystemCollection getMetaSystemCollection() {
      return metaCol;
   }
   
   /**
    * Return whether or not metadata is enabled on this database.
    * @return boolean
    */
   public boolean isMetaEnabled()
   {
      return metaEnabled;
   }

   public Database getDatabase() {
      return this;
   }

   /**
    * flushConfig ensures that the Collection configuration has been
    * properly flushed to disk after a modification.
    */
   public void flushConfig() {
      try {
         Document d = config.getElement().getOwnerDocument();
         sysCol.getCollection(SystemCollection.CONFIGS).setDocument(COLKEY, d);
      }
      catch ( Exception e ) {
         if (log.isErrorEnabled()) {
            log.error("Error Writing Configuration '"+name+"', for database "+getName(), e);
         }
      }
      // observer
      DBObserver.getInstance().flushDatabaseConfig(this, config);
   }

   public boolean close() throws DBException {
      flushConfig();
      databases.remove( getName() );
      return true;
   }

   /**
    * getDocumentCache returns the Database-level Document Cache.
    *
    * @return The DocumentCache
    */
   public DocumentCache getDocumentCache() {
      return docCache;
   }

   /**
    * getQueryEngine returns a reference to the Database's current
    * operating QueryEngine implementation.
    *
    * @return The QueryEngine instance
    */
   public QueryEngine getQueryEngine() {
      return engine;
   }

   // methods for recording the times when meta data is enabled.
   public synchronized void recordTime( String path, long created, long modified )
   {
      TimeRecord rec = (TimeRecord)timestamps.get(path);
      if( null == rec ) {
         rec = new TimeRecord(created, modified);
         timestamps.put(path, rec);
      }
      else
      {
         if( created > 0 )
            rec.setCreatedTime(created);
         if( modified > 0 )
            rec.setModifiedTime(modified);
      }
   }

   public synchronized void removeTime( String path )
   {
      timestamps.remove(path);
   }

   public synchronized TimeRecord getTime( String path )
   {
      TimeRecord rec = (TimeRecord)timestamps.get(path);
      return rec;
   }
}

