/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: FilerTestCase.java,v 1.5 2003/02/06 04:56:39 kstaken Exp $
 */

package org.apache.xindice.core.filer;

import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Record;
import org.apache.xindice.core.data.RecordSet;
import org.apache.xindice.core.data.Value;

import junit.framework.TestCase;

import java.util.List;
import java.util.Vector;

/**
 * @version $Revision: 1.5 $, $Date: 2003/02/06 04:56:39 $
 * @author Kimbro Staken <kstaken@xmldatabases.org>
 * @author Vladimir R. Bossicard <vladimir@apache.org>
 */
public class FilerTestCase
      extends TestCase {

   static public final String TEST_COLLECTION_NAME = "tests";
   static public final Value TEST_VALUE = new Value("<test><test></test><test>Just a test</test></test>");
   static public final Value TEST_VALUE_2 = new Value("<test><test></test><test>Just a test 2</test></test>");
   static public final Value TEST_VALUE_3 = new Value("<test><test></test><test>Just a test 3</test></test>");
   static public final Key TEST_KEY = new Key("test.xml");

   protected Filer filer;

   public void setFiler(Filer filer) {
      this.filer = filer;
   }

   public void setUp()
         throws Exception {
      RecordSet set = filer.getRecordSet();
      while (set.hasMoreRecords()) {
         Key key = set.getNextKey();
         filer.deleteRecord(key);
      }
      assertEquals(0, filer.getRecordCount());
   }

   public String getName() {
      return this.filer.getName();
   }

   public void testSuccessWriteRecord()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      Record result = filer.readRecord(TEST_KEY);
      assertEquals(TEST_VALUE, result.getValue());

      RecordSet set = filer.getRecordSet();
      assertEquals(result.getValue(), set.getNextRecord().getValue());
      assertTrue(set.hasMoreRecords() == false);

      filer.deleteRecord(TEST_KEY);
      assertEquals(0, filer.getRecordCount());
   }

   public void testFailReadDeletedRecord() throws Exception {
      Record result = filer.readRecord(TEST_KEY);
      assertNull(result);
   }

   public void testFailWriteRecordNullKey()
         throws Exception {
      try {
         filer.writeRecord(null, TEST_VALUE);
      } catch (FilerException e) {
         return;
      }
      fail("Expecting FilerException");
   }

   public void testFailWriteRecordNullValue()
         throws Exception {
      try {
         filer.writeRecord(TEST_KEY, null);
      } catch (FilerException e) {
         return;
      }
      fail("Expecting FilerException");
   }

   public void testFailWriteRecordNullValueKey()
         throws Exception {
      try {
         filer.writeRecord(null, null);
      } catch (FilerException e) {
         return;
      }
      fail("Expecting FilerException");
   }

   public void testReadRecord()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      // Valid key
      Record result = filer.readRecord(TEST_KEY);
      assertEquals(TEST_VALUE, result.getValue());
   }

   public void testFailReadRecordNullKey() throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      // Null key
      Record result = filer.readRecord(null);
      assertNull(result);
   }

   public void testFailReadRecordEmptyKey()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      // Empty key
      Record result = filer.readRecord(new Key(""));
      assertNull(result);
   }

   public void testFailReadRecordUnknownKey()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      // Non-existant key
      Record result = filer.readRecord(new Key("non-existant-key"));
      assertNull(result);
   }

   public void testDeleteRecord()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      // Valid key
      filer.deleteRecord(TEST_KEY);
      assertEquals(0, filer.getRecordCount());
      Record result = filer.readRecord(TEST_KEY);
      assertNull(result);
   }

   public void testDeleteRecordNullKey()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      // These should all just fail silently.
      assertTrue(false == filer.deleteRecord(null));
      assertEquals(1, filer.getRecordCount());
   }

   public void testDeleteRecordEmptyKey()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      // These should all just fail silently.
      assertTrue(false == filer.deleteRecord(new Key("")));
      assertEquals(1, filer.getRecordCount());
   }

   public void testDeleteRecordWrongKey()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertEquals(1, filer.getRecordCount());

      // These should all just fail silently.
      assertTrue(false == filer.deleteRecord(new Key("non-existant-key")));
      assertEquals(1, filer.getRecordCount());
   }

   public void testGetRecordCount()
         throws Exception {
      assertEquals(0, filer.getRecordCount());

      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertTrue(filer.writeRecord(new Key("test1"), TEST_VALUE));
      assertTrue(filer.writeRecord(new Key("test2"), TEST_VALUE));
      assertEquals(3, filer.getRecordCount());

      assertTrue(filer.writeRecord(new Key("test3"), TEST_VALUE));
      assertEquals(4, filer.getRecordCount());

      assertTrue(filer.deleteRecord(new Key("test3")));
      assertEquals(3, filer.getRecordCount());

      assertTrue(filer.deleteRecord(TEST_KEY));
      assertTrue(filer.deleteRecord(new Key("test1")));
      assertTrue(filer.deleteRecord(new Key("test2")));

      assertEquals(0, filer.getRecordCount());
   }

   public void testGetRecordSet()
         throws Exception {
      assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
      assertTrue(filer.writeRecord(new Key("test2"), TEST_VALUE_2));
      assertTrue(filer.writeRecord(new Key("test3"), TEST_VALUE_3));

      RecordSet result = filer.getRecordSet();
      assertNotNull(result);

      List results = new Vector();
      while (result.hasMoreRecords()) {
         results.add(result.getNextRecord().getValue());
      }
      assertEquals(3, results.size());
      assertTrue(results.contains(TEST_VALUE));
      assertTrue(results.contains(TEST_VALUE_2));
      assertTrue(results.contains(TEST_VALUE_3));

      assertTrue(filer.deleteRecord(TEST_KEY));
      assertTrue(filer.deleteRecord(new Key("test2")));
      assertTrue(filer.deleteRecord(new Key("test3")));

      result = filer.getRecordSet();
      assertTrue(!result.hasMoreRecords());
   }

   public void testInsertManyDocuments() throws Exception {
      int iterations = 1000;
      for (int i = 0; i < iterations; i++) {
         assertTrue(filer.writeRecord(new Key("key" + i), TEST_VALUE));
      }

      assertTrue(filer.getRecordCount() == iterations);

      for (int i = 0; i < iterations; i++) {
         assertTrue(filer.deleteRecord(new Key("key" + i)));
      }

      assertTrue(filer.getRecordCount() == 0);
   }

}
